//----------------------------------------------------------------------------
//
// ckfb.c - test framebuffer device driver kernel module.
//
//-----------------------------------------------------------------------------

#define MOD_AUTHOR	"Clemens Koller <clemens.koller@anagramm.de>"
#define MOD_DESC	"framebuffer device driver kernel module"

/* 126f  Silicon Motion, Inc. 
   	0501  SM501 VoyagerGX
	0710  SM710 LynxEM 
	0712  SM712 LynxEM+
	0720  SM720 Lynx3DM
	0730  SM731 Cougar3DR
	0810  SM810 LynxE  
	0811  SM811 LynxE  
	0820  SM820 Lynx3D 
	0910  SM910
*/	
	
#define PCI_VENDOR_ID_SILICONMOTION	0x126f
#define PCI_DEVICE_ID_VOYAGER_GX	0x0501

#define MAX_CARDS 4

#include <linux/module.h>	/* Needed by all modules */
#include <linux/kernel.h>	/* Needed for KERN_ALERT */
#include <linux/init.h>		/* Needed for the macros */
#include <linux/pci.h>		/* Needed for all the PCI stuff */
#include "ckfb.h"


static int debug=1;		//insmod parameter
MODULE_PARM(debug,"i");
MODULE_PARM_DESC(debug, "Set debug level [0..1]");

#define dprintk(fmt, args ...) if(debug)printk(KERN_ALERT "ckfb: " fmt, ## args)

/*---------------------------------------------------------------------------*/

// static int ckfb_data __initdata = 3;
struct pci_dev *pdev=NULL;

/*---------------------------------------------------------------------------*/

void dump_struct(char* data, unsigned int size){
	unsigned int i=0;

	dprintk("sizeof: %d\n",size);
	dprintk("ptr: %p\n",data);


	for(i=0;i<10;i++){
		dprintk("%02X\n",*(data+i));


	}
}

/*---------------------------------------------------------------------------*/
//init_module

static int __init ckfb_init(void)
{
	int i=0;
	int err=0;
	
	dprintk("init\n");
	do{
		dprintk("scan %d\n",i);
		
		//0x126f	Silicon Motion Inc.
		//0x0501	SM501 Voyager GX
		pdev=pci_find_device(PCI_VENDOR_ID_SILICONMOTION,
			       PCI_DEVICE_ID_VOYAGER_GX,pdev);
		if(pdev==NULL){
			dprintk("Voyager GX not found! (try: %d)\n",i);
			i++;
		}else{
			dprintk("Voyager GX found!\n");
			dump_struct((char *)pdev,sizeof(struct pci_dev));
			break;
		};
		
	}while(i<MAX_CARDS);

	err=pci_enable_device(pdev);
	if(err){
		dprintk("error: pci_enable_device: %d\n",err);
		return err;
	}

	
	dprintk("No. of devices found: %d\n",i);

	//A non 0 return means init_module failed; module can't be loaded. 
	return 0;
}

/*---------------------------------------------------------------------------*/
//cleanup_module

static void __exit ckfb_exit(void)
{
	pci_disable_device(pdev);
	dprintk("exit\n");
}

/*---------------------------------------------------------------------------*/

module_init(ckfb_init);
module_exit(ckfb_exit)

MODULE_LICENSE("GPL");
MODULE_AUTHOR(MOD_AUTHOR);
MODULE_DESCRIPTION(MOD_DESC);
MODULE_SUPPORTED_DEVICE("fb");

/*************************************************************************/
/* eof */

