/*
 *  linux/drivers/video/voyager.c -- Silicon Motion, Inc. Voyager GX frame buffer device
 *
 *      Copyright (C) 2004 Silicon Motion, Inc.
 *      John Huang, johuang@siliconmotion.com
 *      Ge Wang, gewang@siliconmotion.com
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif

//#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
//#include <linux/errno.h>
//#include <linux/string.h>
//#include <linux/mm.h>
//#include <linux/tty.h>
//#include <linux/slab.h>
//#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/pci.h>
//#include <linux/init.h>
//#include <asm/io.h>

#include "oldvoyager.h"

#define isPCI 1

#define dprintk(fmt, args ...) if(debug)printk(KERN_ALERT "voyager: " fmt, ## args)

/*---------------------------------------------------------------------------*/

#define MOD_AUTHOR      "Clemens Koller <clemens.koller@anagramm.de>"
#define MOD_DESC        "framebuffer device driver kernel module"

/*
 *  126f Silicon Motion, Inc.
 *	0501  SM501 VoyagerGX
 *	0710  SM710 LynxEM 
 *	0712  SM712 LynxEM+
 *	0720  SM720 Lynx3DM
 *	0730  SM731 Cougar3DR
 *	0810  SM810 LynxE
 *	0811  SM811 LynxE
 *	0820  SM820 Lynx3D
 *	0910  SM910
 */

#define PCI_VENDOR_ID_SILICONMOTION     0x126f
#define PCI_DEVICE_ID_VOYAGER_GX        0x0501

static int debug=1;             //insmod parameter
MODULE_PARM(debug,"i");
MODULE_PARM_DESC(debug, "Set debug level [0..1]");

/*---------------------------------------------------------------------------*/
//new structures

/*
static struct _vgxfb_info {
	struct fb_info info;
	struct pci_dev *pdev;
	
	unsigned long mmio_base_phys;
	unsigned long fb_base_phys;
	void *mmio_base;
	void *fb_base;
} vgxfb_info;
*/

static void *VoyagerReg;	// point to virtual Memory Map IO starting address
static void *VoyagerFB;		// point to virtual video memory starting address
static struct fb_info info;	// use fb.h structure
static struct pci_dev *pdev=NULL;	//PCI device structure used in __init and __exit

struct _mode_info {
	unsigned int width;
	unsigned int height;
	unsigned int hz;
} mode_info;			// used to record hardware information

static unsigned long colreg[17];
//static unsigned int readbackerrors=0;
//static unsigned int readbackcounter=0;


//init the fb_var_screeninfo
/*
static struct fb_var_screeninfo vgxfb_var = {
	.xres 		= 640,
	.yres 		= 480,
	.xres_virtual 	= 640,
	.yres_virtual 	= 480,
	.bits_per_pixel = 16,
	.red 		= { 11, 5, 0 },
	.green 		= { 5, 6, 0 }, 
	.blue 		= { 0, 5, 0 },
	.activate	= FB_ACTIVATE_NOW,
	.height		= -1,
	.width		= -1,
	.vmode		= FB_VMODE_NONINTERLACED,
};
*/


static struct fb_var_screeninfo vgxfb_var = {
	.xres 		= 800,
	.yres 		= 600,
	.xres_virtual 	= 800,
	.yres_virtual 	= 600,
	.bits_per_pixel = 32,
	.red 		= { 24, 8, 0 },
	.green 		= {  16, 8, 0 }, 
	.blue 		= {  8, 8, 0 },
	.transp		= {  0, 8, 0 },
	.activate	= FB_ACTIVATE_NOW,
	.height		= -1,
	.width		= -1,
	.vmode		= FB_VMODE_NONINTERLACED,
};


//init a part of the the fb_fix_screnninfo
static struct fb_fix_screeninfo vgxfb_fix = {
	.id 		= "Voyager GX FB",
	.type 		= FB_TYPE_PACKED_PIXELS,
	.visual 	= FB_VISUAL_TRUECOLOR,
	.line_length 	= 800*4,			//length of a line in bytes
	.accel 		= FB_ACCEL_NONE,
};

static struct fb_ops vgxfb_ops = {
	.owner		= THIS_MODULE,
	.fb_setcolreg	= vgxfb_setcolreg,
/*	.fb_fillrect	= cfb_fillrect,*/
/*	.fb_copyarea	= cfb_copyarea,*/
/*	.fb_imageblit	= cfb_imageblit,*/
/*	.fb_cursor	= soft_cursor,*/
};

// Mode table.
mode_table_t mode_table[]={
//----------------------------------------------------------------------------------------
// H.	H.    H.     H.   H.        V.   V.    V.    V.   V.        Pixel     H.     V.
// tot.	disp. sync   sync sync      tot. disp. sync  sync sinc      clock     freq.  freq.
//      end   start  wdth polarity       end   start hght polarity
//----------------------------------------------------------------------------------------
// 640 x 480
{  800, 640,  656,   96,  NEGATIVE, 525, 480,  490,  2,   NEGATIVE, 25175000, 31469, 60 },
{  832, 640,  664,   40,  NEGATIVE, 520, 480,  489,  3,   NEGATIVE, 31500000, 37861, 72 },
{  840, 640,  656,   64,  NEGATIVE, 500, 480,  481,  3,   NEGATIVE, 31500000, 37500, 75 },
{  832, 640,  696,   56,  NEGATIVE, 509, 480,  481,  3,   NEGATIVE, 36000000, 43269, 85 },
// 800 x 600
{ 1024, 800,  824,   72,  POSITIVE, 625, 600,  601,  2,   POSITIVE, 36000000, 35156, 56 },
{ 1056, 800,  840,  128,  POSITIVE, 628, 600,  601,  4,   POSITIVE, 40000000, 37879, 60 },
{ 1040, 800,  856,  120,  POSITIVE, 666, 600,  637,  6,   POSITIVE, 50000000, 48077, 72 },
{ 1056, 800,  816,   80,  POSITIVE, 625, 600,  601,  3,   POSITIVE, 49500000, 46875, 75 },
{ 1048, 800,  832,   64,  POSITIVE, 631, 600,  601,  3,   POSITIVE, 56250000, 53674, 85 },
// 1024 x 768
{ 1344, 1024, 1048, 136,  NEGATIVE, 806, 768,  771,  6,   NEGATIVE, 65000000, 48363, 60 },
{ 1328, 1024, 1048, 136,  NEGATIVE, 806, 768,  771,  6,   NEGATIVE, 75000000, 56476, 70 },
{ 1312, 1024, 1040,  96,  POSITIVE, 800, 768,  769,  3,   POSITIVE, 78750000, 60023, 75 },
{ 1376, 1024, 1072,  96,  POSITIVE, 808, 768,  769,  3,   POSITIVE, 94500000, 68677, 85 },
// End of table.
{ 0, 0, 0, 0, NEGATIVE, 0, 0, 0, 0, NEGATIVE, 0, 0, 0 },
};

//---------------------------------------------------------------------------
//register read/write

unsigned long regRead32(unsigned long nOffset)
{
#if isPCI    
	return swab32(*(volatile unsigned long *)(VoyagerReg+nOffset));
#else
        return *(volatile unsigned long *)(VoyagerReg+nOffset);
#endif
}

void regWrite32(unsigned long nOffset, unsigned long nData)
{
//	unsigned long readbackvalue=0;
#if isPCI    
	*(volatile unsigned long *)(VoyagerReg+nOffset) = swab32(nData);
#else
	*(volatile unsigned long *)(VoyagerReg+nOffset) = nData;
#endif
/*
 	readbackcounter++;
	readbackvalue=regRead32(nOffset);
	if(readbackvalue!=nData) {
		dprintk("readbackerror@ %08lX wr %08lX rd %08lX\n",
			nOffset,nData,readbackvalue);		
		readbackerrors++;
	}
	*/
	
}

/**********************************************************************
 * panelPlaneEnable
 *    Enable panel graphics plane
 **********************************************************************/

void panelPlaneEnable(void)
{
    unsigned long panel_ctrl = 0;

    panel_ctrl = regRead32(PANEL_DISPLAY_CTRL);
    panel_ctrl = FIELD_SET(panel_ctrl, PANEL_DISPLAY_CTRL, PLANE, ENABLE);
    regWrite32(PANEL_DISPLAY_CTRL, panel_ctrl);
}


/**********************************************************************
 * panelPlaneDisable
 *    Disable panel graphics plane
 **********************************************************************/
void panelPlaneDisable(void)
{
    unsigned long panel_ctrl = 0;

    panel_ctrl = regRead32(PANEL_DISPLAY_CTRL);
    panel_ctrl = FIELD_SET(panel_ctrl, PANEL_DISPLAY_CTRL, PLANE, DISABLE);
    regWrite32(PANEL_DISPLAY_CTRL, panel_ctrl);
}


//---------------------------------------------------------------------------
// panelWaitVSync
// Wait for the specified number of panel Vsyncs
// vsync_count - Number of Vsyncs to wait

static void panelWaitVSync(int vsync_count)
{
	unsigned long status;
	unsigned long timeout;

	while (vsync_count-- > 0)
	{
		/* Wait for end of vsync */
		timeout = 0;
		do{
			status = FIELD_GET(regRead32(CMD_INTPR_STATUS),
				CMD_INTPR_STATUS,PANEL_SYNC);
			if (++timeout == VSYNCTIMEOUT) break;
		}while (status == CMD_INTPR_STATUS_PANEL_SYNC_ACTIVE);

		/* Wait for start of vsync */
		timeout = 0;
		do{
			status = FIELD_GET(regRead32(CMD_INTPR_STATUS),
				CMD_INTPR_STATUS,PANEL_SYNC);
			if (++timeout == VSYNCTIMEOUT) break;
		}while (status == CMD_INTPR_STATUS_PANEL_SYNC_INACTIVE);
	}
}


//---------------------------------------------------------------------------
// panelPowerSequence
//    Turn the panel On/Off
//        on_off      - Turn panel On/Off. Can be:
//                      PANEL_ON
//                      PANEL_OFF
//        vsync_delay - Number of Vsyncs to wait after each signal is
//                      turned on/off

static void panelPowerSequence(panel_state_t on_off,int vsync_delay)
{
	unsigned long panelControl = regRead32(PANEL_DISPLAY_CTRL);

	if (on_off == PANEL_ON)
	{
		// Turn on FPVDDEN.
		panelControl = FIELD_SET(panelControl,
			PANEL_DISPLAY_CTRL, FPVDDEN, HIGH);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
		panelWaitVSync(vsync_delay);

		// Turn on FPDATA.
		panelControl = FIELD_SET(panelControl, 
			PANEL_DISPLAY_CTRL, DATA, ENABLE);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
		panelWaitVSync(vsync_delay);

		// Turn on FPVBIAS.
		panelControl = FIELD_SET(panelControl, 
			PANEL_DISPLAY_CTRL, VBIASEN, HIGH);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
		panelWaitVSync(vsync_delay);

		// Turn on FPEN.
		panelControl = FIELD_SET(panelControl, 
			PANEL_DISPLAY_CTRL, FPEN, HIGH);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
	}else{
		// Turn off FPEN.
		panelControl = FIELD_SET(panelControl,
			PANEL_DISPLAY_CTRL, FPEN, LOW);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
		panelWaitVSync(vsync_delay);

		// Turn off FPVBIASEN.
		panelControl = FIELD_SET(panelControl, 
			PANEL_DISPLAY_CTRL, VBIASEN, LOW);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
		panelWaitVSync(vsync_delay);

		// Turn off FPDATA.
		panelControl = FIELD_SET(panelControl, 
			PANEL_DISPLAY_CTRL, DATA, DISABLE);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
		panelWaitVSync(vsync_delay);

		// Turn off FPVDDEN.
		panelControl = FIELD_SET(panelControl, 
			PANEL_DISPLAY_CTRL, FPVDDEN, LOW);
		regWrite32(PANEL_DISPLAY_CTRL, panelControl);
	}
}

/**********************************************************************
 * panelUseCRT
 *    Enable/disable routing of panel output to CRT monitor
 *    [in]
 *        bEnable - TRUE enables routing of panel output to CRT monitor
 *                  FALSE disables routing of panel output to CRT monitor
 **********************************************************************/
void panelUseCRT(int bEnable)
{
	unsigned long panel_ctrl = 0;
	unsigned long crt_ctrl = 0;

	panel_ctrl = regRead32(PANEL_DISPLAY_CTRL);
	crt_ctrl = regRead32(CRT_DISPLAY_CTRL);

	if (bEnable){
		/* Enable panel graphics plane */
		panel_ctrl = FIELD_SET(panel_ctrl, PANEL_DISPLAY_CTRL, PLANE, ENABLE);
		/* Disable CRT graphics plane */
		crt_ctrl = FIELD_SET(crt_ctrl, CRT_DISPLAY_CTRL, PLANE, DISABLE);
		/* Route panel data to CRT monitor */
		crt_ctrl = FIELD_SET(crt_ctrl, CRT_DISPLAY_CTRL, SELECT, PANEL);
	}else{
		/* Disable panel graphics plane */
		panel_ctrl = FIELD_SET(panel_ctrl, PANEL_DISPLAY_CTRL, PLANE, DISABLE);
		/* Enable CRT graphics plane */
		crt_ctrl = FIELD_SET(crt_ctrl, CRT_DISPLAY_CTRL, PLANE, ENABLE);
		/* Do not route panel data to CRT monitor */
		crt_ctrl = FIELD_SET(crt_ctrl, CRT_DISPLAY_CTRL, SELECT, CRT);
	}
	regWrite32(PANEL_DISPLAY_CTRL, panel_ctrl);
	regWrite32(CRT_DISPLAY_CTRL, crt_ctrl);
}


/*---------------------------------------------------------------------------*/
// Program new power mode.
void setPower(unsigned long nGates, unsigned long Clock)
{
	unsigned long gate_reg,clock_reg;
	unsigned long control_value;

	// Get current power mode.
	control_value = FIELD_GET(regRead32(POWER_MODE_CTRL),
			POWER_MODE_CTRL,MODE);

	switch (control_value)
	{
	case POWER_MODE_CTRL_MODE_MODE0:
		// Switch from mode 0 to mode 1.
		gate_reg = POWER_MODE1_GATE;
		clock_reg = POWER_MODE1_CLOCK;
		control_value = FIELD_SET(control_value,
				POWER_MODE_CTRL, MODE, MODE1);
		break;
	case POWER_MODE_CTRL_MODE_MODE1:
	case POWER_MODE_CTRL_MODE_SLEEP:
		// Switch from mode 1 or sleep to mode 0.
		gate_reg = POWER_MODE0_GATE;
		clock_reg = POWER_MODE0_CLOCK;
		control_value = FIELD_SET(control_value, POWER_MODE_CTRL, MODE, MODE0);
		break;
	default:
		// Invalid mode
		return;
	}

	// Program new power mode.
	regWrite32(gate_reg, nGates);
	regWrite32(clock_reg, Clock);
	regWrite32(POWER_MODE_CTRL, control_value);

	// When returning from sleep, wait until finished.
	while (FIELD_GET(regRead32(POWER_MODE_CTRL),POWER_MODE_CTRL,SLEEP_STATUS)
			==POWER_MODE_CTRL_SLEEP_STATUS_ACTIVE) ;
}

/*---------------------------------------------------------------------------*/
// Set DPMS state.
void setDPMS(DPMS_t state)
{
	unsigned long value;

	value = regRead32(SYSTEM_CTRL);
	switch (state)
	{
	case DPMS_ON:
		value = FIELD_SET(value, SYSTEM_CTRL, DPMS, VPHP);
		break;
	case DPMS_STANDBY:
		value = FIELD_SET(value, SYSTEM_CTRL, DPMS, VPHN);
		break;
	case DPMS_SUSPEND:
		value = FIELD_SET(value, SYSTEM_CTRL, DPMS, VNHP);
		break;
	case DPMS_OFF:
		value = FIELD_SET(value, SYSTEM_CTRL, DPMS, VNHN);
		break;
	}
	regWrite32(SYSTEM_CTRL, value);
}


/*---------------------------------------------------------------------------*/
// Perform a rounded division.
long roundDiv(long num, long denom)
{
	/* n / d + 1 / 2 = (2n + d) / 2d */
	return (2 * num + denom) / (2 * denom);
}

/*---------------------------------------------------------------------------*/
// Finds clock closest to the requested.
long findClock(long requested_clock, clock_select_t *clock, display_t display)
{
	long mclk;
	int divider, shift;
	long best_diff = 999999999;

	// Try 288MHz and 336MHz clocks.
	for (mclk = 288000000; mclk <= 336000000; mclk += 48000000)
	{
		// For CRT, try dividers 1 and 3, for panel, try divider 5 as well.
		for (divider = 1; divider <= (display == PANEL ? 5 : 3);
			 divider += 2)
		{
			// Try all 8 shift values.
			for (shift = 0; shift < 8; shift++)
			{
				// Calculate difference with requested clock.
				long diff = roundDiv(mclk, divider << shift) - requested_clock;
				if (diff < 0)
				{
					diff = -diff;
				}

				// If the difference is less than the current, use it.
				if (diff < best_diff)
				{
					// Store best difference.
					best_diff = diff;

					// Store clock values.
					clock->mclk = mclk;
					clock->divider = divider;
					clock->shift = shift;
				}
			}
		}
	}

	// Return best clock.
	return clock->mclk / (clock->divider << clock->shift);
}

/*---------------------------------------------------------------------------*/
// Finds the requested mode in the mode table.
mode_table_t *findMode(mode_table_t *mode_table, int width, int height, int refresh_rate)
{
	// Walk the entire mode table.
	while (mode_table->pixel_clock != 0)
	{
		// If this mode matches the requested mode, return it!
		if ((mode_table->horizontal_display_end == width)
		&& (mode_table->vertical_display_end == height)
		&& (mode_table->vertical_frequency == refresh_rate))
			return(mode_table);

		// Next entry in the mode table.
		mode_table++;
	}

	// No mode found.
	return(NULL);
}

/*---------------------------------------------------------------------------*/
// Converts the VESA timing into Voyager timing.
void adjustMode(mode_table_t *pmode, mode_table_t *mode, display_t display)
{
	long blank_width, sync_start, sync_width;
	clock_select_t clock;

	// Calculate the VESA line and screen frequencies.
	pmode->horizontal_frequency = roundDiv(pmode->pixel_clock,pmode->horizontal_total);
	pmode->vertical_frequency=roundDiv(pmode->horizontal_frequency,pmode->vertical_total);
	// Calculate the sync percentages of the VESA mode.
	blank_width = pmode->horizontal_total-pmode->horizontal_display_end;
	sync_start = roundDiv((pmode->horizontal_sync_start-pmode->horizontal_display_end)
		*100,blank_width);
	sync_width = roundDiv(pmode->horizontal_sync_width*100, blank_width);
	
	// Copy VESA mode into Voyager mode.
	*mode = *pmode;

	// Find the best pixel clock.
	mode->pixel_clock = findClock(pmode->pixel_clock * 2, &clock, display) / 2;

	// Calculate the horizontal total based on the pixel clock and VESA line
	// frequency.
	mode->horizontal_total = roundDiv(mode->pixel_clock,pmode->horizontal_frequency);
	
	// Calculate the sync start and width based on the VESA percentages.
	blank_width = mode->horizontal_total - mode->horizontal_display_end;
	mode->horizontal_sync_start = mode->horizontal_display_end
		+roundDiv(blank_width * sync_start, 100);
	mode->horizontal_sync_width = roundDiv(blank_width * sync_width, 100);

	// Calculate the line and screen frequencies.
	mode->horizontal_frequency = roundDiv(mode->pixel_clock,
		mode->horizontal_total);
	mode->vertical_frequency = roundDiv(mode->horizontal_frequency,
		mode->vertical_total);
}

/*---------------------------------------------------------------------------*/
// Fill the register structure.
void setModeRegisters(reg_table_t *register_table, mode_table_t *mode, display_t display, int bpp)
{
	clock_select_t clock;
	memset(&clock, 0, sizeof(clock));

	// Calculate the clock register values.
	findClock(mode->pixel_clock * 2, &clock, display);

	if (display == PANEL)
	{
		// Set clock value for panel.
		register_table->clock
			= (clock.mclk == 288000000
				? FIELD_SET(0, CURRENT_POWER_CLOCK, P2XCLK_SELECT, 288)
				: FIELD_SET(0, CURRENT_POWER_CLOCK, P2XCLK_SELECT, 336))
			| (clock.divider == 1
				? FIELD_SET(0, CURRENT_POWER_CLOCK, P2XCLK_DIVIDER, 1)
				: (clock.divider == 3
					? FIELD_SET(0, CURRENT_POWER_CLOCK, P2XCLK_DIVIDER, 3)
					: FIELD_SET(0, CURRENT_POWER_CLOCK, P2XCLK_DIVIDER, 5)))
			| FIELD_VALUE(0, CURRENT_POWER_CLOCK, P2XCLK_SHIFT, clock.shift);

		// Set control register value.
		register_table->control
			= (mode->vertical_sync_polarity == POSITIVE
				? FIELD_SET(0, PANEL_DISPLAY_CTRL, VSYNC_PHASE, ACTIVE_HIGH)
				: FIELD_SET(0, PANEL_DISPLAY_CTRL, VSYNC_PHASE, ACTIVE_LOW))
			| (mode->horizontal_sync_polarity == POSITIVE
				? FIELD_SET(0, PANEL_DISPLAY_CTRL, HSYNC_PHASE, ACTIVE_HIGH)
				: FIELD_SET(0, PANEL_DISPLAY_CTRL, HSYNC_PHASE, ACTIVE_LOW))
			| FIELD_SET(0, PANEL_DISPLAY_CTRL, TIMING, ENABLE)
			| FIELD_SET(0, PANEL_DISPLAY_CTRL, PLANE, ENABLE)
			| (bpp == 8
				? FIELD_SET(0, PANEL_DISPLAY_CTRL, FORMAT, 8)
				: (bpp == 16
					? FIELD_SET(0, PANEL_DISPLAY_CTRL, FORMAT, 16)
					: FIELD_SET(0, PANEL_DISPLAY_CTRL, FORMAT, 32)));

		// Set timing registers.
		register_table->horizontal_total
			= FIELD_VALUE(0, PANEL_HORIZONTAL_TOTAL, TOTAL,
						  mode->horizontal_total - 1)
			| FIELD_VALUE(0, PANEL_HORIZONTAL_TOTAL, DISPLAY_END, 
						  mode->horizontal_display_end - 1);

		register_table->horizontal_sync
			= FIELD_VALUE(0, PANEL_HORIZONTAL_SYNC, WIDTH,
						  mode->horizontal_sync_width)
			| FIELD_VALUE(0, PANEL_HORIZONTAL_SYNC, START, 
						  mode->horizontal_sync_start - 1);

		register_table->vertical_total  
			= FIELD_VALUE(0, PANEL_VERTICAL_TOTAL, TOTAL, 
						  mode->vertical_total - 1)
			| FIELD_VALUE(0, PANEL_VERTICAL_TOTAL, DISPLAY_END, 
						  mode->vertical_display_end - 1);

		register_table->vertical_sync    
			= FIELD_VALUE(0, PANEL_VERTICAL_SYNC, HEIGHT, 
						  mode->vertical_sync_height)
			| FIELD_VALUE(0, PANEL_VERTICAL_SYNC, START, 
						  mode->vertical_sync_start - 1);
	}
	else
	{
		// Set clock value for CRT.
		register_table->clock
			= (clock.mclk == 288000000
				? FIELD_SET(0, CURRENT_POWER_CLOCK, V2XCLK_SELECT, 288)
				: FIELD_SET(0, CURRENT_POWER_CLOCK, V2XCLK_SELECT, 336))
			| (clock.divider == 1
				? FIELD_SET(0, CURRENT_POWER_CLOCK, V2XCLK_DIVIDER, 1)
				: FIELD_SET(0, CURRENT_POWER_CLOCK, V2XCLK_DIVIDER, 3))
			| FIELD_VALUE(0, CURRENT_POWER_CLOCK, V2XCLK_SHIFT, clock.shift);

		// Set control register value.
		register_table->control 
			= (mode->vertical_sync_polarity == POSITIVE
				? FIELD_SET(0, CRT_DISPLAY_CTRL, VSYNC_PHASE, ACTIVE_HIGH)
				: FIELD_SET(0, CRT_DISPLAY_CTRL, VSYNC_PHASE, ACTIVE_LOW))
			| (mode->horizontal_sync_polarity == POSITIVE
				? FIELD_SET(0, CRT_DISPLAY_CTRL, HSYNC_PHASE, ACTIVE_HIGH)
				: FIELD_SET(0, CRT_DISPLAY_CTRL, HSYNC_PHASE, ACTIVE_LOW))
			| FIELD_SET(0, CRT_DISPLAY_CTRL, SELECT, CRT)
			| FIELD_SET(0, CRT_DISPLAY_CTRL, TIMING, ENABLE)
			| FIELD_SET(0, CRT_DISPLAY_CTRL, PLANE, ENABLE)
			| (bpp == 8
				? FIELD_SET(0, CRT_DISPLAY_CTRL, FORMAT, 8)
				: (bpp == 16
					? FIELD_SET(0, CRT_DISPLAY_CTRL, FORMAT, 16)
					: FIELD_SET(0, CRT_DISPLAY_CTRL, FORMAT, 32)));

		// Set timing registers.
		register_table->horizontal_total
			= FIELD_VALUE(0, CRT_HORIZONTAL_TOTAL, TOTAL, 
						  mode->horizontal_total - 1)
			| FIELD_VALUE(0, CRT_HORIZONTAL_TOTAL, DISPLAY_END, 
						  mode->horizontal_display_end - 1);

		register_table->horizontal_sync  
			= FIELD_VALUE(0, CRT_HORIZONTAL_SYNC, WIDTH, 
						  mode->horizontal_sync_width)
			| FIELD_VALUE(0, CRT_HORIZONTAL_SYNC, START, 
						  mode->horizontal_sync_start - 1);

		register_table->vertical_total   
			= FIELD_VALUE(0, CRT_VERTICAL_TOTAL, TOTAL, 
						  mode->vertical_total - 1)
			| FIELD_VALUE(0, CRT_VERTICAL_TOTAL, DISPLAY_END, 
						  mode->vertical_display_end - 1);
		register_table->vertical_sync    
			= FIELD_VALUE(0, CRT_VERTICAL_SYNC, HEIGHT, 
						  mode->vertical_sync_height)
			| FIELD_VALUE(0, CRT_VERTICAL_SYNC, START, 
						  mode->vertical_sync_start - 1);
	}

	// Calculate frame buffer width and height.
	register_table->fb_width = mode->horizontal_display_end * (bpp / 8);
	register_table->width = mode->horizontal_display_end;
	register_table->height = mode->vertical_display_end;

	// Save display type.
	register_table->display = display;
}

/*---------------------------------------------------------------------------*/
// Program the mode with the registers specified.
void programMode(reg_table_t *register_table)
{
	unsigned long value, gate, clock;
	unsigned long palette_ram;
	unsigned long fb_size, offset;
	unsigned char red,green,blue;
	unsigned long gray;


	// Get current power configuration.
	gate = regRead32(CURRENT_POWER_GATE);
	clock = regRead32(CURRENT_POWER_CLOCK);

	// Program panel.
	if (register_table->display == PANEL)
	{
		// Program clock, enable display controller.
		gate = FIELD_SET(gate, CURRENT_POWER_GATE, DISPLAY, ENABLE);
		clock &= FIELD_CLEAR(CURRENT_POWER_CLOCK, P2XCLK_SELECT)
			  &  FIELD_CLEAR(CURRENT_POWER_CLOCK, P2XCLK_DIVIDER)
			  &  FIELD_CLEAR(CURRENT_POWER_CLOCK, P2XCLK_SHIFT);
		setPower(gate, clock | register_table->clock);

		// Calculate frame buffer address.
		value = 0;
		fb_size = register_table->fb_width * register_table->height;
		if (FIELD_GET(regRead32(CRT_DISPLAY_CTRL),
			CRT_DISPLAY_CTRL,PLANE)==CRT_DISPLAY_CTRL_PLANE_ENABLE)
		{
			value = FIELD_GET(regRead32(CRT_FB_ADDRESS),
							  CRT_FB_ADDRESS, 
							  ADDRESS);
			if (fb_size < value)
			{
				value = 0;
			}else{
				value += FIELD_GET(regRead32(CRT_FB_WIDTH),
								   CRT_FB_WIDTH,
								   OFFSET)
					  *  (FIELD_GET(regRead32(CRT_VERTICAL_TOTAL),
								   CRT_VERTICAL_TOTAL,
								   DISPLAY_END) + 1);
			}
		}

		// Program panel registers.
		regWrite32(PANEL_FB_ADDRESS,
			FIELD_SET(0, PANEL_FB_ADDRESS, STATUS, PENDING)|
			FIELD_SET(0, PANEL_FB_ADDRESS, EXT, LOCAL)|
		        FIELD_VALUE(0, PANEL_FB_ADDRESS, ADDRESS, value));

		regWrite32(PANEL_FB_WIDTH,
      			FIELD_VALUE(0, PANEL_FB_WIDTH, WIDTH,register_table->fb_width)|
			FIELD_VALUE(0, PANEL_FB_WIDTH, OFFSET,register_table->fb_width));

		regWrite32(PANEL_WINDOW_WIDTH,
			FIELD_VALUE(0, PANEL_WINDOW_WIDTH, WIDTH,register_table->width)|
			FIELD_VALUE(0, PANEL_WINDOW_WIDTH, X, 0));

		regWrite32(PANEL_WINDOW_HEIGHT,
			FIELD_VALUE(0, PANEL_WINDOW_HEIGHT, HEIGHT,register_table->height)|
			FIELD_VALUE(0, PANEL_WINDOW_HEIGHT, Y, 0));
		
		regWrite32(PANEL_PLANE_TL,
			FIELD_VALUE(0, PANEL_PLANE_TL, TOP, 0)|
			FIELD_VALUE(0, PANEL_PLANE_TL, LEFT, 0));

		regWrite32(PANEL_PLANE_BR,
			FIELD_VALUE(0, PANEL_PLANE_BR, BOTTOM,register_table->height - 1)|
			FIELD_VALUE(0, PANEL_PLANE_BR, RIGHT,register_table->width - 1));

		regWrite32(PANEL_HORIZONTAL_TOTAL,register_table->horizontal_total);
		regWrite32(PANEL_HORIZONTAL_SYNC, register_table->horizontal_sync);
		regWrite32(PANEL_VERTICAL_TOTAL,register_table->vertical_total);
		regWrite32(PANEL_VERTICAL_SYNC,register_table->vertical_sync);

		// Program panel display control register.
		value = regRead32(PANEL_DISPLAY_CTRL)
			  & FIELD_CLEAR(PANEL_DISPLAY_CTRL, VSYNC_PHASE)
			  & FIELD_CLEAR(PANEL_DISPLAY_CTRL, HSYNC_PHASE)
			  & FIELD_CLEAR(PANEL_DISPLAY_CTRL, TIMING)
			  & FIELD_CLEAR(PANEL_DISPLAY_CTRL, PLANE)
			  & FIELD_CLEAR(PANEL_DISPLAY_CTRL, FORMAT);

		regWrite32(PANEL_DISPLAY_CTRL, value | register_table->control);

		// Palette RAM.
		palette_ram = PANEL_PALETTE_RAM;

		// Turn on panel.
		panelPowerSequence(PANEL_ON, 4);
	}

	// Program CRT.
	else
	{
		// Program clock, enable display controller.
		gate = FIELD_SET(gate, CURRENT_POWER_GATE, DISPLAY, ENABLE);
		clock &= FIELD_CLEAR(CURRENT_POWER_CLOCK, V2XCLK_SELECT)
			  &  FIELD_CLEAR(CURRENT_POWER_CLOCK, V2XCLK_DIVIDER)
			  &  FIELD_CLEAR(CURRENT_POWER_CLOCK, V2XCLK_SHIFT);
		setPower(gate, clock | register_table->clock);

		// Turn on DAC.
		regWrite32(MISC_CTRL,FIELD_SET(regRead32(MISC_CTRL),MISC_CTRL,DAC_POWER,ENABLE));

		// Calculate frame buffer address.
		value = 0;
		fb_size = register_table->fb_width * register_table->height;
		if (FIELD_GET(regRead32(PANEL_DISPLAY_CTRL),
					  PANEL_DISPLAY_CTRL,
					  PLANE) == PANEL_DISPLAY_CTRL_PLANE_ENABLE)
		{
			value = FIELD_GET(regRead32(PANEL_FB_ADDRESS),
							  PANEL_FB_ADDRESS, 
							  ADDRESS);
			if (fb_size < value)
			{
				value = 0;
			}else{
				value += FIELD_GET(regRead32(PANEL_FB_WIDTH),
								   PANEL_FB_WIDTH,
								   OFFSET)
					  *  FIELD_GET(regRead32(PANEL_WINDOW_HEIGHT),
								   PANEL_WINDOW_HEIGHT,
								   HEIGHT);
			}
		}

		// Program CRT registers.
		regWrite32(CRT_FB_ADDRESS,
			FIELD_SET(0, CRT_FB_ADDRESS, STATUS, PENDING)|
			FIELD_SET(0, CRT_FB_ADDRESS, EXT, LOCAL)|
			FIELD_VALUE(0, CRT_FB_ADDRESS, ADDRESS, value));

		regWrite32(CRT_FB_WIDTH,
			FIELD_VALUE(0, CRT_FB_WIDTH, WIDTH,register_table->fb_width)|
			FIELD_VALUE(0, CRT_FB_WIDTH, OFFSET,register_table->fb_width));

		regWrite32(CRT_HORIZONTAL_TOTAL,register_table->horizontal_total);
		regWrite32(CRT_HORIZONTAL_SYNC,register_table->horizontal_sync);
		regWrite32(CRT_VERTICAL_TOTAL,register_table->vertical_total);
		regWrite32(CRT_VERTICAL_SYNC,register_table->vertical_sync);

		// Program CRT display control register.
		value = regRead32(CRT_DISPLAY_CTRL)
			  & FIELD_CLEAR(CRT_DISPLAY_CTRL, VSYNC_PHASE)
			  & FIELD_CLEAR(CRT_DISPLAY_CTRL, HSYNC_PHASE)
			  & FIELD_CLEAR(CRT_DISPLAY_CTRL, SELECT)
			  & FIELD_CLEAR(CRT_DISPLAY_CTRL, TIMING)
			  & FIELD_CLEAR(CRT_DISPLAY_CTRL, PLANE)
			  & FIELD_CLEAR(CRT_DISPLAY_CTRL, FORMAT);

		regWrite32(CRT_DISPLAY_CTRL,value | register_table->control);

		// Palette RAM.
		palette_ram = CRT_PALETTE_RAM;

		// Turn on CRT.
		setDPMS(DPMS_ON);
	}

	// In case of 8-bpp, fill palette.
	if (FIELD_GET(register_table->control,
				  PANEL_DISPLAY_CTRL,
				  FORMAT) == PANEL_DISPLAY_CTRL_FORMAT_8)
	{
		dprintk("palette fill: 8bpp\n");
		// Start with RGB = 0,0,0.
		red = 0;
		green = 0;
		blue = 0;
		gray = 0;
		for (offset = 0; offset < 256 * 4; offset += 4)
		{
			// Store current RGB value.
			regWrite32(palette_ram + offset, gray
							? RGB((gray + 50) / 100,
								  (gray + 50) / 100,
								  (gray + 50) / 100)
							: RGB(red, green, blue));

			if (gray)
			{
				// Walk through grays (40 in total).
				gray += 654;
			}

			else
			{
				// Walk through colors (6 per base color).
				if (blue != 255)
				{
					blue += 51;
				}
				else if (green != 255)
				{
					blue = 0;
					green += 51;
				}
				else if (red != 255)
				{
					green = blue = 0;
					red += 51;
				}
				else
				{
					gray = 1;
				}
			}
		}
	}

	// For 16- and 32-bpp,  fill palette with gamma values.
	else
	{
		dprintk("pallette fill: 16bpp and 32bpp\n");
		
		// Start with RGB = 0,0,0.
		value = 0x000000;
		for (offset = 0; offset < 256 * 4; offset += 4)
		{
			regWrite32(palette_ram + offset, value);

			// Advance RGB by 1,1,1.
			value += 0x010101;
		}
	}
}

/*---------------------------------------------------------------------------*/

void SetMode(unsigned long nWidth, unsigned long nHeight,
	unsigned long nHertz, display_t display, int bpp)
{
	mode_table_t mode;
	pmode_table_t pmode;
	reg_table_t register_table;

	// Locate the mode
	pmode = findMode(mode_table, nWidth, nHeight, nHertz);
	if (pmode != NULL)
	{
		dprintk("Vesa mode ok.\n");
		// Convert VESA timing into Voyager timing
		adjustMode(pmode, &mode, display);

		// Fill the register structure
		setModeRegisters(&register_table, &mode, display, bpp);

		// Program the registers
		programMode(&register_table);
	}else{
		dprintk("Vesa mode not found!\n");
	}
	
}

/*---------------------------------------------------------------------------*/
//functions called from FBDEV...


static int vgxfb_setcolreg(unsigned regno, unsigned red, unsigned green,
	unsigned blue, unsigned transp, struct fb_info *info)
{
	if (regno>15) return 1;
	((u16 *)(info->pseudo_palette))[regno]=
		(red & 0xf800) | (green & 0xfc00 >> 5) | (blue & 0xf800 >> 11);
	return 0;
}

/*---------------------------------------------------------------------------*/

static int __init voyagerfb_init(void)
{
	int err=0;
//	unsigned long i=0;
	unsigned int x,y;
//	unsigned int dummy;
	unsigned char testcol;
	
	unsigned long pFramebufferPhysical;
	unsigned long pRegPhysical;

	dprintk("init called\n");
	
	//find our card
	pdev=pci_find_device(PCI_VENDOR_ID_SILICONMOTION,
			PCI_DEVICE_ID_VOYAGER_GX, pdev);
	if(pdev==NULL){
		dprintk("Voyager GX not found!\n");
		return -ENODEV;
	}else{
		dprintk("Voyager GX found!\n");
	}
	err=pci_enable_device(pdev); // enable voyager chip
	if(err){
		dprintk("error: pci_enable_device: %d\n",err);       
		return err;
	}

	//clear fb info structure	
	memset(&info, 0, sizeof(struct fb_info));

	vgxfb_fix.smem_start = pci_resource_start(pdev, 0);
	vgxfb_fix.smem_len = pci_resource_len(pdev,0);

	info.flags = FBINFO_FLAG_DEFAULT;
	info.fbops = &vgxfb_ops;
	info.var = vgxfb_var;
	info.fix = vgxfb_fix;
	info.pseudo_palette = colreg;

	//map address and memory stuff
	pFramebufferPhysical=pci_resource_start(pdev,0);
	dprintk("pFramebufferPhysical: %04lX\n",pFramebufferPhysical);
	pRegPhysical=pci_resource_start(pdev,1);
	dprintk("pRegPhysical        : %04lX\n",pRegPhysical);
	
	//ioremap
	VoyagerFB=ioremap(pFramebufferPhysical,0x00800000);
	if(!VoyagerFB){
	 	dprintk("FB memory failed to remap\n");
		return -ENOMEM;
	}
	dprintk("VoyagerFB           : %p\n",VoyagerFB);
	
	VoyagerReg=ioremap(pRegPhysical,0x00200000);
	if(!VoyagerReg){
	 	dprintk("MMIO failed to remap\n");
		return -ENOMEM;
	}
	dprintk("VoyagerReg          : %p\n",VoyagerReg);

	info.screen_base=VoyagerFB;
	mode_info.width=info.var.xres;
	mode_info.height=info.var.yres;
	mode_info.hz=60;

	SetMode(info.var.xres,info.var.yres,mode_info.hz,PANEL,
		info.var.bits_per_pixel);
	//Turn on CRT DAC (for simultanous display only)
	regWrite32(MISC_CTRL, FIELD_SET(regRead32(MISC_CTRL),
		MISC_CTRL,DAC_POWER,ENABLE));
		
	err=register_framebuffer(&info);
	if(err){
		dprintk("error: register_framebuffer failed - err: %d",err);
		return err;
	}
	
	//dprintk("SystemControl       : %08lX\n",regRead32(SYSTEM_CTRL));
	//dprintk("MiscControl         : %08lX\n",regRead32(MISC_CTRL));
	//dprintk("DRAMControl         : %08lX\n",regRead32(DRAM_CTRL));
	//dprintk("ArbitrationControl  : %08lX\n",regRead32(ARBITRATION_CTRL));
/*
	dprintk("System Configuration:\n");
	for(i=0;i<0x70;i+=4){
		dprintk("%08lX = %08lX\n",i,regRead32(i));
	}

	dprintk("Display Controller:\n");
	for(i=0x80000;i<0x80400;i+=4){
		dprintk("%08lX = %08lX\n",i,regRead32(i));
	}
*/	
	for(y=0;y<info.var.yres;y++){
	for(x=0;x<info.var.xres;x++){
		testcol=(x*0xff/info.var.xres)/2+(y*0xff/info.var.xres)/2;
	/*	if(x%4==0) testcol=0x7e;
		if(x%4==1) testcol=0x7f;
		if(x%4==2) testcol=0x80;
		if(x%4==3) testcol=0x81;
	*/	
		*(char*)(VoyagerFB+((y*info.var.xres+x)*4)+0)=testcol;	//R
		*(char*)(VoyagerFB+((y*info.var.xres+x)*4)+1)=testcol;	//G
		*(char*)(VoyagerFB+((y*info.var.xres+x)*4)+2)=testcol;	//B
		*(char*)(VoyagerFB+((y*info.var.xres+x)*4)+3)=0x00;	//alpha
	}
	}
/*
	dummy=0;	
	for(y=0;y<info.var.yres;y++){
	for(x=0;x<info.var.xres;x++){
		testcol=(x*0xff/info.var.xres)/2+(y*0xff/info.var.xres)/2;
		if(*(char*)(VoyagerFB+((y*info.var.xres+x)*4)+0)!=testcol) dummy++;	//R
		if(*(char*)(VoyagerFB+((y*info.var.xres+x)*4)+1)!=testcol) dummy++;	//G
		if(*(char*)(VoyagerFB+((y*info.var.xres+x)*4)+2)!=testcol) dummy++;	//B
		if(*(char*)(VoyagerFB+((y*info.var.xres+x)*4)+3)!=0x00) dummy++;	//alpha
	}
	}
*/
//	dprintk("memory r/w errors: %d\n",dummy);
//	dprintk("readback errors: %d\n",readbackerrors);
//	dprintk("readback counter: %d\n",readbackcounter);
	dprintk("init complete\n");
	return 0;
}

/*---------------------------------------------------------------------------*/
//cleanup module

static void __exit voyagerfb_exit(void)
{
	int err=0;
	
	dprintk("exit called\n");

	//Turn off CRT DAC (for simultanous display only)
	regWrite32(MISC_CTRL, FIELD_SET(regRead32(MISC_CTRL),
		MISC_CTRL,DAC_POWER,DISABLE));
	panelPowerSequence(PANEL_OFF,4);
	
	//call kernel to release me	
	err=unregister_framebuffer(&info);
	if(err){
		dprintk("error: unregister_framebuffer failed - err: %d",err);
	}else{
		dprintk("VoyagerFB           : %p\n",VoyagerFB);
		iounmap(VoyagerFB);
		dprintk("VoyagerReg          : %p\n",VoyagerReg);
		iounmap(VoyagerReg);
		pci_disable_device(pdev);
	}
	dprintk("exit complete\n");
}

/*---------------------------------------------------------------------------*/

module_init(voyagerfb_init);
module_exit(voyagerfb_exit)

MODULE_LICENSE("GPL");
MODULE_AUTHOR(MOD_AUTHOR);
MODULE_DESCRIPTION(MOD_DESC);
MODULE_SUPPORTED_DEVICE("fb");


/*****************************************************************************/
/* eof */

